from django.conf import settings
from django.utils.importlib import import_module
from django.utils.module_loading import module_has_submodule

def autodiscover():
    """
    Automatically import the events module for all INSTALLED_APPS.  Based on
    ``django.contrib.admin.autodiscover``.
    """
    
    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        try:
            import_module("%s.events" % app)
        except ImportError:
            if module_has_submodule(mod, "events"):
                raise



from django.db import models

class StreamItem(models.Model):
    context = models.TextField()
    remove = models.BooleanField()
    clusters = models.ManyToManyField("StreamCluster", related_name="items")

class StreamCluster(models.Model):
    event_type = models.CharField(max_length=64)
    clustered_on = models.CharField(max_length=64)



import hashlib
import time
import uuid
from collections import defaultdict, namedtuple
from datetime import datetime, timedelta

import redis

from django.conf import settings
from django.db.models import Model
from django.template import Context
from django.template.loader import render_to_string
from django.utils import simplejson as json

from .models import (StreamItem as StreamItemModel,
    StreamCluster as StreamClusterModel)

# TODO: ...
def get_redis_connection():
    return redis.Redis(**settings.REDIS_SETTINGS)

class EventTypeMetaclass(type):
    def __new__(cls, name, bases, attrs):
        new_cls = super(EventTypeMetaclass, cls).__new__(cls, name, bases, attrs)
        if hasattr(new_cls, "context_shape"):
            for k, v in new_cls.context_shape.iteritems():
                if issubclass(v, Model):
                    class Klass(ModelContextItemType):
                        model = v
                    new_cls.context_shape[k] = Klass
        if hasattr(new_cls, "slug"):
            new_cls.registry[new_cls.slug] = new_cls
            assert new_cls.slug is not None
            assert new_cls.context_shape is not None
            assert new_cls.queryable_by is not None
            assert new_cls.default_cluster_by is not None

        return new_cls

class EventType(object):
    __metaclass__ = EventTypeMetaclass

    registry = {}

    cluster = True

    def __init__(self, context, timestamp=None, remove=False):
        if timestamp is None:
            timestamp = datetime.now()
        self.timestamp = timestamp
        for key, spec in self.context_shape.iteritems():
            if key not in context:
                raise ValueError("Missing value from context")
            if not spec.valid_obj(context[key]):
                raise TypeError("Invalid context item for %s: %s" % (key, context[key]))
        self.context = context
        self.remove = remove

    def serialize_context(self, context):
        result = {}
        for key, spec in self.context_shape.iteritems():
            result[key] = spec.serialize(context[key])
        return result

    @property
    def redis(self):
        if not hasattr(self, "_redis"):
            self._redis = get_redis_connection()
        return self._redis

    def save(self):
        context = self.serialize_context(self.context)
        s = StreamItemModel.objects.create(
            context = json.dumps(context),
            remove = self.remove,
        )

        t = self.timestamp
        timestamp = time.mktime(t.timetuple()) + 1e-6 * t.microsecond

        record = {
            "id": s.pk,
            "context": self.serialize_context(self.context),
            "remove": self.remove,
            "timestamp": tuple(t.timetuple())[:-3],
        }
        for field in self.queryable_by:
            obj_key = self.context_shape[field](self.context[field]).lookup_key()
            keys = [obj_key, "%s:%s" % (obj_key, self.slug)]
            c = None
            for key in keys:
                c = self._add_to_key(field, key, timestamp, record, c, s)
        self._add_to_key(
            self.default_cluster_by, "ALL_EVENTS", timestamp, record, c, s
        )

    def _add_to_key(self, field, key, timestamp, record, c, s):
        for item, score in self.redis.zrevrange(key, 0, 5, withscores=True):
            cluster_timestamp = datetime.fromtimestamp(score)
            data = json.loads(item)
            if (data["slug"] == self.slug and self.cluster and
                data["items"][0]["context"][data["clustered_on"]] == self.context_shape[field].serialize(self.context[field]) and
                self.timestamp - cluster_timestamp < timedelta(minutes=5)):
                c = StreamClusterModel.objects.get(pk=data["cluster_id"])
                c.items.add(s)
                data["items"].append(record)
                self.redis.zrem(key, item)
                self.redis.zadd(key, json.dumps(data), score)
                break
        else:
            if c is None:
                c = StreamClusterModel.objects.create(
                    event_type = self.slug,
                    clustered_on = field,
                )
                c.items.add(s)
            data = json.dumps({
                "slug": self.slug,
                "items": [record],
                "clustered_on": field,
                "cluster_id": c.pk,
            })
            self.redis.zadd(key, data, timestamp)
        return c

class ContextItemType(object):
    def __init__(self, obj):
        self.obj = obj

    def lookup_key(self):
        return self.obj

    @classmethod
    def unique_key(cls):
        return cls

    @classmethod
    def valid_obj(cls, obj):
        return True

    @classmethod
    def serialize(cls, obj):
        return obj

    @classmethod
    def deserialize(cls, obj):
        return obj

    @classmethod
    def deserialize_bulk(cls, objs):
        return dict(
            (obj, cls.deserialize(obj))
            for obj in objs
        )

class ModelContextItemType(ContextItemType):
    model = None

    def lookup_key(self):
        return "%s:%s:%s" % (
            self.model._meta.app_label,
            self.model._meta.object_name,
            self.obj.pk
        )

    @classmethod
    def unique_key(cls):
        return cls.model

    @classmethod
    def valid_obj(cls, obj):
        return isinstance(obj, cls.model) or isinstance(obj, (int, long))

    @classmethod
    def serialize(cls, obj):
        return obj.pk

    @classmethod
    def deserialize(cls, obj):
        return cls.model._default_manager.get(pk=obj)

    @classmethod
    def deserialize_bulk(cls, objs):
        return cls.model._default_manager.in_bulk(objs)

class StreamItem(object):
    def __init__(self, slug, timestamp, context, item_id, cluster_id):
        self.slug = slug
        self.timestamp = timestamp
        self.context = context
        self.item_id = item_id
        self.cluster_id = cluster_id

    def __getattr__(self, name):
        return self.context[name]

class StreamCluster(object):
    def __init__(self, slug, date_added, events, clustered_on=None,
        cluster_id=None):
        self.slug = slug
        self.date_added = date_added
        self.events = events
        self.clustered_on = clustered_on
        self.cluster_id = cluster_id

    def __iter__(self):
        return iter(self.events)

    def __len__(self):
        return len(self.events)

    def __unicode__(self):
        return self.render()

    @property
    def date_updated(self):
        return max(e.timestamp for e in self.events)

    def render(self, context=None):
        if context is None:
            context = Context()
        context.push()
        context["query_object"] = self.clustered_on
        if len(self.events) == 1:
            context["event"] = self.events[0]
            # The context variable for a full stream is often named events,
            # since we're providing the full context, set this to None so the
            # event template isn't confused (since events means something
            # different).  This is safe since we push and pop the context
            # before and after modifying it.
            context["events"] = None
        else:
            context["events"] = self
            # Same reason as above.
            context["event"] = None
        try:
            return render_to_string("events/event/%s.html" % self.slug, context)
        finally:
            context.pop()

RawResults = namedtuple("RawResults", ["field", "vals"])
Status = namedtuple("Status", ["adds", "removes"])

class Stream(object):
    def __init__(self, *objs, **kwargs):
        event_type = kwargs.pop("event_type", None)
        limit = kwargs.pop("limit", 20)
        offset = kwargs.pop("offset", 0)

        if kwargs:
            raise TypeError("Unexpected keyword argument: %s" % kwargs)

        final_objs = []
        for obj in objs:
            if isinstance(obj, Model):
                class Klass(ModelContextItemType):
                    model = type(obj)
                final_objs.append(Klass(obj))
            else:
                final_objs.append(obj)

        self.objs = final_objs
        self.event_type = event_type
        self.limit = limit
        self.offset = offset

    def __iter__(self):
        redis = get_redis_connection()

        postfix = ""
        if self.event_type is not None:
            postfix += ":%s" % (self.event_type.slug)
        lookup_keys = [
            obj.lookup_key() + postfix
            for obj in self.objs
        ]

        if len(lookup_keys) >= 2:
            s = hashlib.sha1()
            for lookup_key in lookup_keys:
                s.update(lookup_key)
            key = s.hexdigest()
            redis.zunionstore(key, lookup_keys, aggregate="MIN")
            # Expire it in 5 minutes, enough that paginating shouldn't require
            # a recompute, but short enough to not clutter the place up.
            redis.expire(key, 60 * 5)
        elif len(lookup_keys) == 1:
            key = lookup_keys[0]
        else:
            assert not self.event_type
            key = "ALL_EVENTS"

        statuses = defaultdict(lambda: Status(0, 0))
        items = list(redis.zrevrange(key, self.offset, self.limit, withscores=True))
        parsed_items = []
        context_items = {}
        for cluster, score in items:
            data = json.loads(cluster)
            parsed_items.append((data, score))
            for o in data["items"]:
                status_key = self._status_key(data["slug"], o)
                status = statuses[status_key]
                if o["remove"]:
                    statuses[status_key] = status._replace(removes=status.removes+1)
                else:
                    statuses[status_key] = status._replace(adds=status.adds+1)
                for key, val in o["context"].iteritems():
                    field = EventType.registry[data["slug"]].context_shape[key]
                    key = field.unique_key()
                    if key not in context_items:
                        context_items[key] = RawResults(field, set())
                    context_items[key].vals.add(val)

        final_context_items = {}
        for key, (field, vals) in context_items.iteritems():
            final_context_items[key] = field.deserialize_bulk(vals)

        for data, score in parsed_items:
            cluster_items = []
            timestamp = datetime.fromtimestamp(score)
            for o in data["items"]:
                item = self._convert_item(
                    data["slug"], o, timestamp, statuses, final_context_items, data["cluster_id"]
                )
                if item is not None:
                    cluster_items.append(item)
            if cluster_items:
                clustered_on = None
                if data["clustered_on"] is not None:
                    clustered_on = cluster_items[0].context[data["clustered_on"]]
                yield StreamCluster(
                    data["slug"],
                    timestamp,
                    cluster_items,
                    clustered_on,
                    data["cluster_id"]
                )

    def _convert_item(self, slug, data, timestamp, statuses, context_items,
        cluster_id):
        status_key = self._status_key(slug, data)
        status = statuses[status_key]
        if data["remove"]:
            current_attr, other_attr = "removes", "adds"
        else:
            current_attr, other_attr = "adds", "removes"

        if getattr(status, current_attr) <= getattr(status, other_attr):
            return
        statuses[status_key] = status._replace(**{
            current_attr: getattr(status, current_attr) - 1
        })

        context = {}
        for key, value in data["context"].iteritems():
            field = EventType.registry[slug].context_shape[key]
            context[key] = context_items[field.unique_key()][value]

        return StreamItem(
            slug,
            datetime(*data["timestamp"]),
            context,
            data["id"],
            cluster_id,
        )

    def _status_key(self, slug, data):
        return (
            slug,
            tuple(sorted(data["context"].items()))
        )



from django import template

from templatetag_sugar.register import tag
from templatetag_sugar.parser import Variable

register = template.Library()

@tag(register, [Variable("event")])
def render_event(context, event):
    return event.render(context)



from contextlib import contextmanager
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth.models import User as UserModel
from django.db import connection, connections, DEFAULT_DB_ALIAS
from django.template import TemplateDoesNotExist
from django.test import TestCase

from .base import get_redis_connection, EventType, ContextItemType, Stream, StreamCluster
from .models import StreamItem, StreamCluster as StreamClusterModel

class EventTestCase(TestCase):
    @contextmanager
    def assert_raises(self, error_type):
        exc_info = {}
        try:
            yield exc_info
        except error_type, e:
            exc_info["exception"] = e
        except Exception, e:
            self.fail("Exception of type %s expected, %s gotten" % (error_type, e))
        else:
            self.fail("Exception of type %s expected, but not raised" % error_type)


class User(ContextItemType):
    @classmethod
    def valid_obj(cls, obj):
        return isinstance(obj, basestring)


class Follow(EventType):
    slug = "follow"
    context_shape = {
        "follower": User,
        "following": User,
    }
    queryable_by = ["follower", "following"]
    default_cluster_by = "follower"

class Poke(EventType):
    slug = "poke"
    context_shape = {
        "poker": User,
        "pokee": User,
    }
    queryable_by = ["poker", "pokee"]
    default_cluster_by = "poker"

class SomeEvent(EventType):
    slug = "some-event"
    context_shape = {
        "user": UserModel
    }
    queryable_by = ["user"]
    default_cluster_by = "user"

class AnotherEvent(EventType):
    slug = "another-event"
    context_shape = {
        "user": UserModel
    }
    queryable_by = ["user"]
    default_cluster_by = "user"

class Review(EventType):
    slug = "review"
    context_shape = {
        "reviewer": User,
    }
    queryable_by = ["reviewer"]
    cluster = False
    default_cluster_by = "reviewer"

_missing = object()

class EventTests(EventTestCase):
    def setUp(self):
        self.original_REDIS_SETTINGS = getattr(settings, "REDIS_SETTINGS", _missing)
        settings.REDIS_SETTINGS = {
            "db": 9,
        }

    def tearDown(self):
        get_redis_connection().flushdb()
        if self.original_REDIS_SETTINGS is _missing:
            del settings.REDIS_SETTINGS
        else:
            settings.REDIS_SETTINGS = self.original_REDIS_SETTINGS

    def assert_stream_equal(self, stream, objs):
        got = list(stream)
        self.assertEqual(len(got), len(objs))
        for cluster, expected in zip(got, objs):
            self.assertEqual(len(cluster), len(expected))
            self.assertEqual(cluster.slug, expected.slug)
            self.assertEqual(cluster.date_added, expected.date_added)
            self.assertEqual(cluster.date_updated, expected.date_updated)
            # This isn't relevant to most tests, so we make it optional.
            if expected.clustered_on is not None:
                self.assertEqual(cluster.clustered_on, expected.clustered_on)
            for res, obj in zip(cluster, expected):
                self.assertEqual(res.timestamp, obj.timestamp)
                self.assertEqual(res.context, obj.context)
                self.assertEqual(res.slug, obj.slug)


    def test_context_item_coercion(self):
        with self.assert_raises(ValueError):
            Follow({
                "follower": "alex",
            }, timestamp=datetime(2010, 10, 8))
        with self.assert_raises(TypeError):
            Follow({
                "following": "alex",
                "follower": 2
            }, timestamp=datetime(2010, 10, 8))

    def test_event_save(self):
        event = Follow({
            "follower": "alex",
            "following": "einstein",
        })
        event.save()

        self.assertEqual(StreamItem.objects.count(), 1)
        # One for the follower, one for the following,
        self.assertEqual(StreamClusterModel.objects.count(), 2)
        s = Stream(User("alex"))
        c = iter(s).next()
        self.assertEqual(
            c.cluster_id,
            StreamClusterModel.objects.get(clustered_on="follower").pk
        )

    def test_event_stream_single(self):
        event = Follow({
            "following": "alex",
            "follower": "einstein",
        }, timestamp=datetime(2010, 10, 8))
        event.save()

        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("follow", datetime(2010, 10, 8), [
                Follow({
                    "following": "alex",
                    "follower": "einstein",
                }, datetime(2010, 10, 8)),
            ])
        ])

    def test_event_stream_multiple(self):
        event = Follow({
            "following": "alex",
            "follower": "jacob",
        }, timestamp=datetime(2010, 10, 8, 12, 30, 12))
        event.save()

        self.assert_stream_equal(Stream(User("alex"), User("jacob")), [
            StreamCluster("follow", datetime(2010, 10, 8, 12, 30, 12), [
                Follow({
                    "following": "alex",
                    "follower": "jacob",
                }, datetime(2010, 10, 8, 12, 30, 12)),
            ]),
            StreamCluster("follow", datetime(2010, 10, 8, 12, 30, 12), [
                Follow({
                    "following": "alex",
                    "follower": "jacob",
                }, datetime(2010, 10, 8, 12, 30, 12)),
            ]),
        ])

    def test_cluster(self):
        Follow({
            "following": "alex",
            "follower": "jacob",
        }, timestamp=datetime(2010, 10, 8, 12, 30)).save()
        Follow({
            "following": "alex",
            "follower": "daniel",
        }, timestamp=datetime(2010, 10, 8, 12, 31)).save()
        Follow({
            "following": "alex",
            "follower": "james",
        }, timestamp=datetime(2010, 10, 8, 13, 30)).save()

        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("follow", datetime(2010, 10, 8, 13, 30), [
                Follow({
                    "following": "alex",
                    "follower": "james",
                }, datetime(2010, 10, 8, 13, 30))
            ]),
            StreamCluster("follow", datetime(2010, 10, 8, 12, 30), [
                Follow({
                    "following": "alex",
                    "follower": "jacob",
                }, datetime(2010, 10, 8, 12, 30)),
                Follow({
                    "following": "alex",
                    "follower": "daniel",
                }, datetime(2010, 10, 8, 12, 31)),
            ]),
        ])

    def test_model(self):
        u = UserModel.objects.create_user("joe", "joe@schmoe.net", "abc123")
        d = datetime(2010, 10, 21, 15, 56, 22)
        SomeEvent({
            "user": u,
        }, timestamp=d).save()

        self.assert_stream_equal(Stream(u), [
            StreamCluster("some-event", d, [
                SomeEvent({
                    "user": u,
                }, d)
            ])
        ])

    def test_model_clustered_on(self):
        u = UserModel.objects.create_user("joe", "joe@schmoe.net", "abc123")
        d = datetime(2010, 10, 21, 15, 56, 22, 330000)
        SomeEvent({
            "user": u,
        }, timestamp=d).save()

        s = Stream(u)
        c = iter(s).next()
        self.assertEqual(c.clustered_on, u)

    def test_render_cluster(self):
        Follow({
            "following": "alex",
            "follower": "daniel",
        }).save()
        Follow({
            "following": "alex",
            "follower": "aaron",
        }).save()
        s = Stream(User("alex"))
        f = iter(s).next()

        with self.assert_raises(TemplateDoesNotExist) as exc_info:
            f.render()
        self.assertEqual(exc_info["exception"].args, ("events/event/follow.html",))

    def test_remove(self):
        d1 = datetime(2010, 10, 8, 9, 32)
        d2 = datetime(2010, 10, 8, 9, 30)
        c = {
            "following": "alex",
            "follower": "daniel",
        }
        Follow(c, d1).save()
        Follow(c, d2, remove=True).save()

        self.assert_stream_equal(Stream(User("alex")), [])

    # TODO: Design decision
    # def test_remove_first(self):
    #     '''
    #     A case when event is removed, but there was no paired event before that.
    #     This can happend if event for following was not created before unfollowing
    #     '''
    #     d1 = datetime(2010, 10, 8, 9, 32)
    #     d2 = datetime(2010, 10, 8, 9, 30)
    #     c = {
    #         "following": "alex",
    #         "follower": "daniel",
    #     }
    #     Follow(c, d1, remove=True).save()
    #     # Follow(c, d2).save()
    # 
    #     self.assert_stream_equal(Stream(User("alex")), [])

    def test_remove_cluster(self):
        d1 = datetime(2010, 10, 8, 9, 30)
        d2 = datetime(2010, 10, 8, 9, 32)
        c = {
            "following": "alex",
            "follower": "daniel",
        }
        Follow(c, d1).save()
        Follow(c, d1, remove=True).save()
        Follow({
            "following": "alex",
            "follower": "jacob",
        }, d2).save()

        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("follow", d1, [
                Follow({
                    "following": "alex",
                    "follower": "jacob",
                }, d2)
            ]),
        ])

    def test_cluster_types(self):
        d1 = datetime(2010, 10, 8, 9, 32)
        d2 = datetime(2010, 10, 8, 9, 33)
        fc1 = {
            "following": "alex",
            "follower": "daniel"
        }
        fc2 = {
            "following": "alex",
            "follower": "ryan"
        }
        pc1 = {
            "pokee": "alex",
            "poker": "michael"
        }
        pc2 = {
            "pokee": "alex",
            "poker": "ralph"
        }
        Follow(fc1, d1).save()
        Follow(fc2, d2).save()
        Poke(pc1, d1).save()
        Poke(pc2, d2).save()

        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("poke", d1, [
                Poke(pc1, d1),
                Poke(pc2, d2),
            ]),
            StreamCluster("follow", d1, [
                Follow(fc1, d1),
                Follow(fc2, d2),
            ])
        ])

    def test_unpaired_remove(self):
        d = datetime(2010, 10, 8, 9, 32)
        c = {
            "following": "alex",
            "follower": "daniel"
        }
        Follow(c, d, remove=True).save()
        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("follow", d, [
                Follow(c, d, remove=True)
            ]),
        ])

    def test_clustered_on(self):
        d = datetime(2010, 10, 8, 9, 32)
        c1 = {
            "following": "alex",
            "follower": "daniel"
        }
        Follow(c1, d).save()
        s = Stream(User("alex"))
        c = iter(s).next()
        self.assertEqual(c.clustered_on, "alex")

    def test_no_cluster(self):
        d = datetime(2010, 10, 8, 9, 32)
        c = {"reviewer": "Chris"}
        Review(c, d).save()
        Review(c, d).save()
        self.assert_stream_equal(Stream(User("Chris")), [
            StreamCluster("review", d, [
                Review(c, d),
            ]),
            StreamCluster("review", d, [
                Review(c, d),
            ]),
        ])

    def test_efficient_context_deserialize(self):
        d1 = datetime(2010, 10, 8, 9, 32)
        d2 = datetime(2010, 10, 8, 9, 33)
        u1 = UserModel.objects.create_user("joe", "joe@schmoe.net", "abc123")
        u2 = UserModel.objects.create_user("bob", "bob@schmoe.net", "123abc")
        SomeEvent({"user": u1}, d1).save()
        SomeEvent({"user": u2}, d2).save()

        # 1 query to get all of the users, that's it.
        with self.assertNumQueries(1):
            list(Stream(u1, u2))

    def test_efficient_context_deserialiez_different_events(self):
        d1 = datetime(2010, 10, 8, 9, 32)
        d2 = datetime(2010, 10, 8, 9, 33)
        u1 = UserModel.objects.create_user("joe", "joe@schmoe.net", "abc123")
        u2 = UserModel.objects.create_user("bob", "bob@schmoe.net", "123abc")
        SomeEvent({"user": u1}, d1).save()
        AnotherEvent({"user": u2}, d2).save()

        # 1 query to get all of the users, even though they're on different
        # event types.
        with self.assertNumQueries(1):
            list(Stream(u1, u2))

    def test_multiple_create_remove(self):
        c = {
            "follower": "alex",
            "following": "daniel"
        }
        ds = [
            datetime(2010, 10, 8, 12) + timedelta(minutes=1) * i for i in xrange(4)
        ]
        for d in ds:
            Follow(c, d).save()
            Follow(c, d + timedelta(seconds=30), remove=True).save()
        Follow(c, datetime(2010, 10, 8, 12, 4, 30)).save()

        self.assert_stream_equal(Stream(User("alex")), [
            StreamCluster("follow", datetime(2010, 10, 8, 12), [
                Follow(c, datetime(2010, 10, 8, 12)),
            ])
        ])

    def test_cluster_db(self):
        c = {
            "follower": "alex",
            "following": "daniel"
        }
        ds = [
            datetime(2010, 10, 8, 12) + timedelta(minutes=1) * i for i in xrange(4)
        ]
        for d in ds:
            Follow(c, d).save()

        self.assertEqual(StreamClusterModel.objects.count(), 2)
        s = StreamClusterModel.objects.get(clustered_on="follower")
        self.assertEqual(s.event_type, "follow")
        self.assertEqual(s.items.count(), 4)

    def test_all_stream(self):
        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)

        c1 = {
            "follower": "alex",
            "following": "daniel"
        }
        c2 = {
            "follower": "aaron",
            "following": "charlie",
        }

        Follow(c1, d1).save()
        Follow(c2, d2).save()

        self.assert_stream_equal(Stream(), [
            StreamCluster("follow", d2, [
                Follow(c2, d2),
            ], clustered_on="aaron"),
            StreamCluster("follow", d1, [
                Follow(c1, d1),
            ], clustered_on="alex")
        ])

    def test_shared_key(self):
        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)
        c1 = {
            "follower": "alex",
            "following": "daniel"
        }
        c2 = {
            "follower": "aaron",
            "following": "charlie",
        }
        Follow(c1, d1).save()
        Follow(c2, d2).save()

        redis = get_redis_connection()
        # 1 - ALL_EVENTS
        # 8 - each username + each username:follow
        # 9
        self.assertEqual(len(redis.keys()), 9)

        list(Stream(User("alex"), User("aaron")))
        self.assertEqual(len(redis.keys()), 10)
        list(Stream(User("alex"), User("aaron")))
        self.assertEqual(len(redis.keys()), 10)

    def test_offset(self):
        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)
        c1 = {
            "follower": "alex",
            "following": "daniel"
        }
        c2 = {
            "follower": "aaron",
            "following": "charlie",
        }
        Follow(c1, d1).save()
        Follow(c2, d2).save()

        self.assert_stream_equal(Stream(offset=1), [
            StreamCluster("follow", d1, [
                Follow(c1, d1)
            ])
        ])

    def test_cluster_by_model(self):
        u = UserModel.objects.create_user("me", "hi", "me@me.com")

        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)
        c = {"user": u}
        SomeEvent(c, d1).save()
        SomeEvent(c, d2).save()

        self.assert_stream_equal(Stream(u), [
            StreamCluster("some-event", d1, [
                SomeEvent(c, d1),
                SomeEvent(c, d2),
            ])
        ])

    def test_cluster_id(self):
        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)
        c1 = {
            "follower": "alex",
            "following": "daniel"
        }
        c2 = {
            "follower": "alex",
            "following": "aaron",
        }
        Follow(c1, d1).save()

        s = iter(Stream(User("alex"))).next()
        self.assertEqual(
            s.cluster_id,
            StreamClusterModel.objects.get(clustered_on="follower").pk
        )
        self.assertEqual(s.events[0].cluster_id, s.cluster_id)

        Follow(c2, d2).save()
        s = iter(Stream(User("alex"))).next()
        self.assertEqual(
            s.cluster_id,
            StreamClusterModel.objects.get(clustered_on="follower").pk
        )
        self.assertEqual(s.events[0].cluster_id, s.cluster_id)
        self.assertEqual(s.events[1].cluster_id, s.cluster_id)

    def test_times(self):
        d1 = datetime(2010, 10, 8, 12, 30)
        d2 = datetime(2010, 10, 8, 12, 33)
        c1 = {
            "follower": "alex",
            "following": "daniel"
        }
        c2 = {
            "follower": "alex",
            "following": "aaron",
        }
        Follow(c1, d1).save()
        Follow(c2, d2).save()

        s = iter(Stream(User("alex"))).next()
        self.assertEqual(s.date_added, d1)
        self.assertEqual(s.date_updated, d2)



from setuptools import setup, find_packages
 
setup(
    name='django-timeline',
    version='0.5',
    description='a Django timeline (activity stream) using redis',
    author='Chris Drackett',
    author_email='chris@tiltshiftstudio.com',
    url = "https://github.com/tiltshift/django-timeline.git",
    packages = find_packages(),
    include_package_data=True,
    install_requires=[
        'django>=1.3.1',
        'redis>=2.0.0'
    ],
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Framework :: Django",
    ]
)